package de.hamster.editor.view;

import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;

import javax.swing.BorderFactory;
import javax.swing.JTextPane;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Element;
import javax.swing.undo.UndoManager;

import de.hamster.compiler.model.JavaToken;
import de.hamster.editor.controller.EditorController;
import de.hamster.model.HamsterFile;
import de.hamster.workbench.Utils;

/**
 * Diese Klasse stellt eine einzelne TextArea dar, die sich in einem
 * Registerreiter der TabbedTextArea befindet. Sie unterstuetzt
 * Syntax-Highlighting, Undo, Redo und die Hervorhebung von einzelnen Zeilen.
 * 
 * @author $Author: djasper $
 * @version $Revision: 1.6 $
 */
public class TextArea extends JTextPane
		implements
			UndoableEditListener,
			DocumentListener,
			Printable {

	/**
	 * Dieses Attribut aktiviert das Highlighting der aktuellen Zeile.
	 */
	private boolean activeLineHighlighted;

	protected long lastModified;

	protected int line;

	/**
	 * Diese Datei wird aktuell editiert.
	 */
	protected HamsterFile file;

	/**
	 * Der UndoManager dieses Dokuments.
	 */
	protected UndoManager undoManager;

	protected TabbedTextArea tabbedTextArea;

	public TextArea(TabbedTextArea tabbedTextArea, EditorController controller,
			HamsterFile file) {
		super(new JavaDocument(false));
		this.tabbedTextArea = tabbedTextArea;
		this.file = file;
		load();
		setCaretPosition(0);
		undoManager = new UndoManager();
		JavaDocument document = (JavaDocument) getDocument();
		document.addUndoableEditListener(this);
		document.addDocumentListener(this);
		line = -1;
	}

	public TextArea(String text) {
		super(new JavaDocument(true));
		setText(text);
		setBorder(BorderFactory.createEmptyBorder());
		JavaDocument document = (JavaDocument) getDocument();
		document.rehighlight(0, text.length());
		line = -1;
	}
	
	// Martin
	public TextArea(TabbedTextArea tabbedTextArea, EditorController controller,
			HamsterFile file, boolean scheme) {
		super(new SchemeDocument(false));
		this.tabbedTextArea = tabbedTextArea;
		this.file = file;
		load();
		setCaretPosition(0);
		undoManager = new UndoManager();
		SchemeDocument document = (SchemeDocument) getDocument();
		document.addUndoableEditListener(this);
		document.addDocumentListener(this);
		line = -1;
	}

	// Martin
	public TextArea(String text, boolean scheme) {
		super(new SchemeDocument(true));
		setText(text);
		setBorder(BorderFactory.createEmptyBorder());
		SchemeDocument document = (SchemeDocument) getDocument();
		document.rehighlight(0, text.length());
		line = -1;
	}

	public void refreshFile() {
		if (getParent() == null)
			return;
		if (file.lastModified() == 0 && lastModified != 0)
			tabbedTextArea.closeFile(file);
		// TODO: Workaround: Speichern und direkt danach das Aenderungsdatum
		// lesen funktioniert nicht
		else if (Math.abs(lastModified - file.lastModified()) > 100) {
			System.out.println(lastModified + " != " + file.lastModified());
			if (Utils.ask(this, "editor.dialog.modifiedexternal")) {
				load();
			} else {
				lastModified = file.lastModified();
			}
		}
	}

	/**
	 * Diese Methode liefert den undoManager
	 * 
	 * @return undoManager
	 */
	public UndoManager getUndoManager() {
		return undoManager;
	}

	/**
	 * Diese Methode bringt das Caret in eine andere Zeile.
	 * 
	 * @param line
	 *            Die neue Zeile des Carets
	 */
	public void markLine(int line) {
		Element root = getDocument().getDefaultRootElement();
		setCaretPosition(root.getElement(line).getStartOffset());
		this.line = line;
		repaint();
	}

	/**
	 * Mit dieser Methode wird ein Fehler ausgewaehlt.
	 * 
	 * @param line
	 *            Die Zeile in der sich der Fehler befindet
	 * @param column
	 *            Die Spalte, an der der Fehler beginnt
	 */
	public void selectError(int line, int column) {
		Element root = getDocument().getDefaultRootElement();
		int pos = root.getElement(line).getStartOffset();
		String text = getText();
		for (int i = 0; i < column; i++) {
			if (text.charAt(pos + i) == '\t')
				column -= 7;
		}
		JavaDocument document = (JavaDocument) getDocument();
		JavaToken token = document.getTokenAt(pos + column);
		if (token != null) {
			setSelectionStart(token.getStart());
			setSelectionEnd(token.getStart() + token.getText().length());
		}
	}

	// Martin
	public void undoableEditHappened(UndoableEditEvent e) {
		if(file.getType() == HamsterFile.SCHEMEPROGRAM) {
			if (!((SchemeDocument) getDocument()).isHighlighting()) {
				undoManager.addEdit(e.getEdit());
			}
		} else {
			if (!((JavaDocument) getDocument()).isHighlighting()) {
				undoManager.addEdit(e.getEdit());
			}
		}
	}

	/**
	 * Diese Methode liefert die Datei, die gerade editiert wird
	 * 
	 * @return Die aktuell editierte Datei.
	 */
	public HamsterFile getFile() {
		return file;
	}

	public void save() {
		//System.out.println(1);
		file.save(getText());
		//System.out.println(2);
		lastModified = file.lastModified();
		//System.out.println(3);
	}
	public void load() {
		setText(file.load());
		file.setModified(false);
		lastModified = file.lastModified();
	}

	/**
	 * Diese Methode setzt die aktuell verwendete Datei
	 * 
	 * @param file
	 *            Die neue Datei
	 */
	public void setFile(HamsterFile file) {
		this.file = file;
		lastModified = file.lastModified();
	}

	public void changedUpdate(DocumentEvent e) {
	}

	public void insertUpdate(DocumentEvent e) {
		file.setModified(true);
	}
	public void removeUpdate(DocumentEvent e) {
		file.setModified(true);
	}

	public void paint(Graphics g) {
		if (line != -1) {
			int y = 0;
			try {
				Element root = getDocument().getDefaultRootElement();
				y = (int) modelToView(root.getElement(line).getStartOffset())
						.getY();
			} catch (BadLocationException e) {
			}
			g.setColor(getBackground());
			g.fillRect(0, 0, getWidth(), getHeight());
			g.setColor(getSelectionColor());
			g.fillRect(0, y, getWidth(), 17);
			setOpaque(false);
			super.paint(g);
			setOpaque(true);
		} else {
			super.paint(g);
		}
	}
	public void removeLineHighlight() {
		line = -1;
	}

	public int print(Graphics graphics, PageFormat pageFormat, int pageIndex)
			throws PrinterException {
		double length = getSize().getHeight();
		double pageHeight = (int) pageFormat.getImageableHeight();
		double pageWidth = (int) pageFormat.getImageableWidth();
		TextArea t = new TextArea(getText());
		t.setSize((int) pageWidth, 100000);
		if (pageIndex <= length / pageHeight) {
			Graphics2D g = (Graphics2D) graphics
					.create((int) pageFormat.getImageableX(), (int) (pageFormat
							.getImageableY() - pageIndex * pageHeight),
							(int) pageWidth, (int) length);

			t.print(g);
			return Printable.PAGE_EXISTS;
		} else {
			return Printable.NO_SUCH_PAGE;
		}
	}
}